//*******************************************************************
//*******************************************************************
//***  Implementation of the CoDetails component object
//***  This component is an inproc component that is used for
//***  memory management and performance exercises.
//*******************************************************************
//*******************************************************************

//*******************************************************************
//** Note that this code is an exact duplica of the out-of-process
//** server implementation.  All the code for the component object
//** and its factory is the same.  The differences are in the
//** server lifetime management code, server registration code,
//** and DLL entry points.
//*******************************************************************

//-------------------------------------------------------------------
// Required macros
//-------------------------------------------------------------------
#define UNICODE         // UNICODE
#define _WIN32_DCOM     // DCOM

//-------------------------------------------------------------------
// includes
//-------------------------------------------------------------------
#include <assert.h>
#include <stdio.h>
#include <windows.h>
#include <initguid.h>
#include "..\idl\details.h"

//-------------------------------------------------------------------
// global event telling use wen to shutdown our application
//-------------------------------------------------------------------
HANDLE g_hExitEvent;
wchar_t g_wszModuleName[MAX_PATH];

// Number of living objects
LONG g_lComponentRefCounts = 0; // counts living objects and locks

//-------------------------------------------------------------------
// Component Level Reference count for the lifetime managment of 
// the whole component.
// These specific implementations are used for in-process servers
// There are a pair of functions like these that are used for
// out-of-process servers.
// Return values for debugging purpose only 
//-------------------------------------------------------------------
inline ULONG ComponentAddRef()
{
   return (InterlockedIncrement(&g_lComponentRefCounts));
}

inline ULONG ComponentRelease()  
{
   return (InterlockedDecrement(&g_lComponentRefCounts));
}

//*******************************************************************
//*******************************************************************
//***  UTILITY FUNCTIONS
//*******************************************************************
//*******************************************************************
void DisplayStatus(wchar_t *pwszMsg, HRESULT hr)
{

   if (hr == S_OK) {
      wprintf(TEXT("%s\n"), pwszMsg);
      return;
   }

   if (HRESULT_FACILITY(hr) == FACILITY_WINDOWS) {
      hr = HRESULT_CODE(hr);
   }

   wchar_t *pwszStatus;
   FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
       FORMAT_MESSAGE_FROM_SYSTEM,
       NULL,
       hr,
       MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
       (LPWSTR)&pwszStatus,
       0,
       NULL );

   wprintf(TEXT("%s: %s (ECode: %lx)\n"), pwszMsg, pwszStatus, hr);

   LocalFree(pwszStatus);
}

//-------------------------------------------------------------------
// Given a unicode string, get the number of bytes needed to host it
//-------------------------------------------------------------------
inline long ByteLen(wchar_t *pwsz) 
{
   return (sizeof(wchar_t)*(wcslen(pwsz)+1));
}

//-------------------------------------------------------------------
// Automatic registration; for in-process server, do:
//    regsvr32.exe xxx.dll
//-------------------------------------------------------------------
void RegisterServerToSystem()
{
   wchar_t wszKey[MAX_PATH];
   wchar_t wszValue[MAX_PATH];
   HKEY hKey = 0;

   // CLSID_Details = {3CC346E1-D3CC-11d1-915C-006008052F2D}
   // HKEY_CLASSES_ROOT\CLSID\{3CC346E1-D3CC-11d1-915C-006008052F2D}
   //      @="Details"
   wcscpy(wszKey, TEXT("CLSID\\{3CC346E1-D3CC-11d1-915C-006008052F2D}"));
   RegCreateKey(HKEY_CLASSES_ROOT, wszKey, &hKey);
   wcscpy(wszValue, TEXT("Details"));
   RegSetValueEx(hKey, 0, 0, REG_SZ, (BYTE*)wszValue, ByteLen(wszValue));
   RegCloseKey(hKey);

   // HKEY_CLASSES_ROOT\CLSID\{3CC346E1-D3CC-11d1-915C-006008052F2D}\InprocServer32
   //      @="E:\DCOM\detailsDLL.dll"
   //      "ThreadingModel" = "Both"
   wcscpy(wszKey, TEXT("CLSID\\{3CC346E1-D3CC-11d1-915C-006008052F2D}")
        TEXT("\\InprocServer32"));
   RegCreateKey(HKEY_CLASSES_ROOT, wszKey, &hKey);
   wcscpy(wszValue, g_wszModuleName);
   RegSetValueEx(hKey, 0, 0, REG_SZ, (BYTE*)wszValue, ByteLen(wszValue));
   wcscpy(wszValue, TEXT("Both"));
   RegSetValueEx(hKey, TEXT("ThreadingModel"), 0, REG_SZ, 
                 (BYTE*)wszValue, ByteLen(wszValue));
   RegCloseKey(hKey);
}

//-------------------------------------------------------------------
// Automatic registration; for in-process server, do:
//    regsvr32.exe -u xxx.dll
//-------------------------------------------------------------------
void UnregisterServerFromSystem()
{
   long lRc = 0 ;

   lRc=RegDeleteKey(HKEY_CLASSES_ROOT, 
       TEXT("CLSID\\{3CC346E1-D3CC-11d1-915C-006008052F2D}") 
     TEXT("\\InprocServer32"));
   DisplayStatus(TEXT("Unregistered CLSID's InprocServer"), lRc);

   lRc=RegDeleteKey(HKEY_CLASSES_ROOT, 
       TEXT("CLSID\\{3CC346E1-D3CC-11d1-915C-006008052F2D}"));
   DisplayStatus(TEXT("Unregistered CLSID"), lRc);
}
 
//*******************************************************************
//*******************************************************************
//***  Class CoDetails - A component object that implements 
//***  interface:  IDetails 
//***  Simple implementation that doesn't support aggregation.
//***  This object cannot be aggregated by another object,
//***  Containment works with every component object!
//*******************************************************************
//*******************************************************************
class CoDetails : public IDetails {
public:
   // constructors/destructors
   CoDetails() : m_lRefCount(0) { ComponentAddRef(); }
   ~CoDetails() { ComponentRelease(); }

   // The factory will call this method to actually 
   // create this object
   static HRESULT CreateObject(LPUNKNOWN pUnkOuter, REFIID riid, 
                               void** ppv);
public: 
   // IUnknown Methods
   STDMETHODIMP QueryInterface(REFIID riid, void **ppv);
   STDMETHODIMP_(ULONG) AddRef(void)
   {
      //------debug messages
      // wchar_t wszMsg[MAX_PATH];
      // wsprintf(wszMsg, TEXT("CoDetails::AddRef (%ld)"), m_lRefCount); 
      // DisplayStatus(wszMsg, S_OK);
      return InterlockedIncrement(&m_lRefCount);
   }
   STDMETHODIMP_(ULONG) Release(void)
   {
      //------debug messages
      // wchar_t wszMsg[MAX_PATH];
      // wsprintf(wszMsg, TEXT("CoDetails::Release (%ld)"), m_lRefCount-1); 
      // DisplayStatus(wszMsg, S_OK);

      long lCount = InterlockedDecrement(&m_lRefCount);
      if (lCount == 0) {
         // DisplayStatus(TEXT("CoDetails::Release - delete object"), S_OK);
         delete this; 
      } 
      return lCount; 
   }

   // IDetails Methods
   STDMETHODIMP VoidCall(void);
   STDMETHODIMP LongCall(long lNumber);
   STDMETHODIMP ArrayCall(long lSize, long *pArrayLong);

private:
   LONG m_lRefCount;
};

//-------------------------------------------------------------------
// CoDetails::CreateObject - static function to create an
// CoDetails object
//-------------------------------------------------------------------
HRESULT CoDetails::CreateObject(LPUNKNOWN pUnkOuter, 
                               REFIID riid, 
                               void** ppv)
{
   *ppv = NULL;

   // CoDetails doesn't support aggregation
   if (pUnkOuter != NULL) { return CLASS_E_NOAGGREGATION; }

   CoDetails * pDetails = new CoDetails;
   if (pDetails == NULL) { return E_OUTOFMEMORY; }

   HRESULT hr = pDetails->QueryInterface(riid, ppv);
   if (FAILED(hr)) { delete pDetails; }

   return hr;
}

//-------------------------------------------------------------------
// CoDetails::QueryInterface
//-------------------------------------------------------------------
STDMETHODIMP
CoDetails::QueryInterface(REFIID riid, void** ppv)
{
   //DisplayStatus(TEXT("CoDetails::QueryInterface"), S_OK);

   if (ppv==NULL) { return E_INVALIDARG; }

   if (riid==IID_IUnknown) {
       *ppv= static_cast<IDetails *>(this);
   } else if (riid==IID_IDetails) {
       *ppv= static_cast<IDetails *>(this);
   } else {
       *ppv=NULL; return E_NOINTERFACE ;
   }

   reinterpret_cast<IUnknown *>(*ppv)->AddRef();

   return S_OK;
}

//-------------------------------------------------------------------
// CoDetails::VoidCall
// For performance testing....
//-------------------------------------------------------------------
STDMETHODIMP CoDetails::VoidCall(void)
{
   return S_OK;
}

//-------------------------------------------------------------------
// CoDetails::LongCall
// For performance testing....
//-------------------------------------------------------------------
STDMETHODIMP CoDetails::LongCall(long lNumber)
{
   return S_OK;
}

//-------------------------------------------------------------------
// CoDetails::ArrayCall
// For performance testing....
//-------------------------------------------------------------------
STDMETHODIMP CoDetails::ArrayCall(long lSize, long *pArrayLong)
{
   return S_OK;
}

//*******************************************************************
//*******************************************************************
//***  Class Factory: manufacturing CoDetails objects
//*******************************************************************
//*******************************************************************
class CoDetailsFactory : public IClassFactory 
{
public:
   // IUnknown Methods
   STDMETHODIMP QueryInterface (REFIID riid, void** ppv);
   STDMETHODIMP_(ULONG) AddRef(void)
   { return 1; }
   STDMETHODIMP_(ULONG) Release(void) 
   { return 1; }

   // IClassFactory Methods
   STDMETHODIMP CreateInstance(LPUNKNOWN pUnkOuter, 
                               REFIID riid, 
                               void **ppv)
  {
     // call CoDetails's static function to create
     // a CoDetails component object
     return CoDetails::CreateObject(pUnkOuter, riid, ppv);
  }

   STDMETHODIMP LockServer(BOOL fLock)
   {
      if (fLock) {
          ComponentAddRef();
      } else {
          ComponentRelease();
      }
      return S_OK;        
   }   
};

//-------------------------------------------------------------------
//  Singleton factory instance that manufacture CoCorDetails
//  component objects
//-------------------------------------------------------------------
CoDetailsFactory g_DetailsClassFactory;

//-------------------------------------------------------------------
//  CoDetailsFactory::QueryInterface
//-------------------------------------------------------------------
STDMETHODIMP 
CoDetailsFactory::QueryInterface(REFIID riid, void** ppv)
{
   if (ppv==NULL) { return E_INVALIDARG; }

   if (riid==IID_IUnknown) {
      *ppv= static_cast<IClassFactory *>(this);
   } else if (riid==IID_IClassFactory) {
      *ppv= static_cast<IClassFactory *>(this);
   } else {
      *ppv=NULL; return E_NOINTERFACE ;
   }

   reinterpret_cast<IUnknown *>(*ppv)->AddRef();

   return S_OK;
}

//*******************************************************************
//*******************************************************************
//***  DLL Exports
//*******************************************************************
//*******************************************************************
extern "C" BOOL WINAPI DllMain(HINSTANCE hInstance, 
                               DWORD dwReason, 
                               LPVOID /*lpReserved*/)
{
   if (dwReason == DLL_PROCESS_ATTACH)
   {
        GetModuleFileName(hInstance, g_wszModuleName, MAX_PATH);
   }
   else if (dwReason == DLL_PROCESS_DETACH)
   {
   }
   return TRUE;    // ok
}

//*******************************************************************
//*******************************************************************
//** Used to determine whether the DLL can be unloaded by DCOM
//*******************************************************************
//*******************************************************************
STDAPI DllCanUnloadNow(void)
{ 
   if (g_lComponentRefCounts ==0) {
      return S_OK;
   } else {
      return S_FALSE;
   }
}

//*******************************************************************
//*******************************************************************
//** Returns a class factory to create an object of the requested type
//*******************************************************************
//*******************************************************************
STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID* ppv)
{
   *ppv = NULL;

   // Does our component support this particular class ID?
   if (CLSID_Details!=rclsid) { return E_FAIL; }

   CoDetailsFactory * pCF = new CoDetailsFactory();
   if (pCF==NULL) { return E_OUTOFMEMORY; }

   HRESULT hr = pCF->QueryInterface(riid, ppv);
   if (FAILED(hr)) {
      delete pCF;
   } else {
      *ppv=pCF;
   }

   return hr;
}

//*******************************************************************
//*******************************************************************
//**  DllRegisterServer - Adds entries to the system registry
//*******************************************************************
//*******************************************************************
STDAPI DllRegisterServer(void)
{
   RegisterServerToSystem();
   return S_OK;
}

//*******************************************************************
//*******************************************************************
//** DllUnregisterServer - Removes entries from the system registry
//*******************************************************************
//*******************************************************************
STDAPI DllUnregisterServer(void)
{
   UnregisterServerFromSystem();
   return S_OK;
}
